//  *********************************************************************
//  * DISCLAIMER                                                        *
//  *                                                                   *
//  * Neither the authors of this software system, nor their employing  *
//  * institutes, nor the agencies providing financial support for this *
//  * work  make  any representation or  warranty, express or implied,  *
//  * regarding  this  software system or assume any liability for its  *
//  * use.                                                              *
//  *                                                                   *
//  * This  code  implementation is the  intellectual property  of the  *
//  * OpenGATE collaboration.                                           *
//  * By copying,  distributing  or modifying the Program (or any work  *
//  * based  on  the Program)  you indicate  your  acceptance of  this  *
//  * statement, and all its terms.                                     *
//  *********************************************************************
//
//######################################################################################
//# Authors   : Sadek A. Nehmeh, CR Schmidtlein (v1.0)                                 #
//#              JJM Kouwenberg (v1.1)                                                 # 
//#                                                                                    #
//# Program   : Bin_GATE_v1.1.c  03-MAR-2015                                           # 
//#                                                                                    #   
//# Objective : To read the coincidences TTree from the .root file, and generates the  #
//#              corresponding Michelogram and Projection files.                       #
//#                                                                                    #
//# Input     : Monte Carlo data from GATE and egsPET                                  #
//#                                                                                    #
//# Output    : 1 Michelogram files according to various binning definitions           #
//#           : 2 Projection  files according to various binning definitions           #
//#                                                                                    #
//# HOW TO COMPILE:                                                                    #
//# 1) Compile using this command line in the terminal:                                #
//#       g++ Bin_GATE_v1.1.c `root-config --cflags --libs` -o Bin_GATE.out            #
//#                                                                                    #
//# HOW TO RUN:                                                                        #
//# 1) After compiling the code type the following ([] means not optional):            #
//#       ./Bin_GATE.out 'directory name of root files' 'output file name' ...         #
//#            'max ring difference' ['scan ID' ['Random substraction (0,1)'           #
//#                    ['trues (0,1)' 'randoms (0,1)' 'scatters (0,1)']]]              #
//#                                                                                    #
//# CHANGELOG                                                                          #
//#  v1.1: Dynamic data support, delay window random correction support,               #
//#       extra arguments added and code was compressed                                #
//######################################################################################
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "TROOT.h"
#include "TSystem.h"
#include "TChain.h"
#include "TH2D.h" 
#include "TDirectory.h"
#include "TList.h"
#include "Rtypes.h"
#include "TChainElement.h"
#include "TTree.h"
#include "TFile.h"
#include "TStyle.h"
#include "TH2.h"
#include "TH2F.h"
#include "TCanvas.h"
#include "TRandom.h"

using namespace std ;

// DLS Parameters
#define   N_RINGS                        52 // DLS: Number of rings
#define   N_SEG                          103 // DLS: Number of segments (2 N_RINGS - 1)
#define   N_DET                          624 // DLS: Detectors per ring
#define   S_WIDTH                        312 // DLS: Number of radial sinogram bins
#define   N_RSEC                         48 // DLS: Number of resector
#define   N_MODULE                       4 // DLS: Number of modules (1x4)
#define   N_MOD_xy                       1 // DLS: Number of tangential modules
#define   N_MOD_z                        4 // DLS: Number of axial modules
#define   N_SUBMOD                       1 // DLS: Number of submodules (1x1)
#define   N_SMOD_xy                      1 // DLS: Number of tangential submodules
#define   N_SMOD_z                       1 // DLS: Number of axial submodules
#define   N_CRYSTAL                      169 // DLS: Number of crystals (13x13)
#define   N_CRY_xy                       13 // DLS: Number of tangential crystals
#define   N_CRY_z                        13 // DLS: Number of axial crystals
#define   USE_OFFSET                     0 // DLS: On/Off use of offset
#define   OFFSET                         0 // DLS: Sets initial sinogram angle

unsigned short n_ring_dif, b_delay, b_trues, b_randoms, b_scatters;
short scan_id;
/*
   Current data constructions available for output.
    Michelograms: [ring1][ring2][phi][u]
*/

Float_t Mich_r1r2fu[N_RINGS][N_RINGS][N_DET/2][S_WIDTH]={0}; 

int main(const int argc, const char** argv)
{
   //---------------------------------------------------------------------------------
   // the first argument (argv[1]) is the sub-directory of the input file
   // the second argument (argv[2]) is the name of the output file
   // the third argument (argv[3]) is the number of the maximum ring difference
   // the fourth argument (argv[4]) is the scan ID that should be converted from the ROOT file.
   //   Default: -1, indicates specific ID so that all scans will be converted
   // the 5 - 7th arguments (argv[5 - 7]) indicate whether respectively the trues, 
   // randoms and scatters should be incorporated. 1 for true, 0 for false.
   //   Default: 1
   //---------------------------------------------------------------------------------
   if(argc<4) {
      std::cout<<"Insufficient number of arguments provided!"<<std::endl ;
      return 1;
   } else if (argc == 4) {
      scan_id      = -1;
      b_delay      = 0;
      b_trues      = 1;
      b_randoms    = 1;
      b_scatters   = 1;
   } else if (argc == 5) {
      scan_id      = atoi(argv[4]);
      b_delay      = 0;
      b_trues      = 1;
      b_randoms    = 1;
      b_scatters   = 1;
   } else if (argc == 6) {
      scan_id      = atoi(argv[4]);
      b_delay      = atoi(argv[5]);
      b_trues      = 1;
      b_randoms    = 1;
      b_scatters   = 1;
   } else {
      scan_id      = atoi(argv[4]);
      b_delay      = atoi(argv[5]);
      b_trues      = atoi(argv[6]);
      b_randoms    = atoi(argv[7]);
      b_scatters   = atoi(argv[8]);
   }
   
   Double_t PI   = acos(-1.0);
   Int_t nbytes  = 0;
   
   n_ring_dif   = atoi(argv[3]);

   string   filedir, inputfilename;
   string   filename, Moutputfilename, Poutputfilename;
   FILE *Mich_r1r2fuFile, *Proj_File;
   
   filedir = argv[1] ; 
   inputfilename = filedir + "*.root" ;
   
   cout << "Input file name is " << inputfilename << endl;
   
   filename = argv[2];
   if (scan_id > -1) filename = filename + argv[4];
   Moutputfilename = "Mich_"+ filename + ".s" ;
   Poutputfilename = filename + ".s" ;      
   Mich_r1r2fuFile = fopen(Moutputfilename.c_str(),"wb");
   Proj_File       = fopen(Poutputfilename.c_str(),"wb");
   
   //####################################################################
   //#   Declaration of leaves types - TTree Coincidences               #
   //####################################################################
   
   Float_t    axialPos, rotationAngle, sinogramS, sinogramTheta;
   Char_t     comptVolName1[40], comptVolName2[40];
   Int_t      compton1, compton2;
   Int_t      runID, sourceID1, sourceID2, eventID1, eventID2; 
   Int_t      layerID1, layerID2, crystalID1, crystalID2;
   Int_t      submoduleID1, submoduleID2, moduleID1, moduleID2, rsectorID1, rsectorID2;
   Int_t      comptonPhantom1, comptonPhantom2;
   Float_t    energy1, energy2;    
   Float_t    globalPosX1, globalPosX2, globalPosY1, globalPosY2, globalPosZ1, globalPosZ2;
   Float_t    sourcePosX1, sourcePosX2, sourcePosY1, sourcePosY2, sourcePosZ1, sourcePosZ2;
   Double_t   time1, time2;
   
   short      random_true;
   Int_t      ring1, ring2, crystal1, crystal2;
   Int_t      phi, u;
   Int_t      Counts = 0;   
   int        zi, c1, c2;
   Int_t      nentries;
   TChain     *Coincidences;
   
   //######################################################################################
   //#   Set branch addresses - TTree Coincidences
   //#    t_cd=0 Coincidences
   //#    t_cd=1 Delay window coincidences, if b_delay = 1
   //######################################################################################
   
   for (short t_cd=0; t_cd<(1+b_delay); t_cd++)
   {
      if (t_cd == 1)
      {
         Coincidences = new TChain("delay");
         Coincidences->Add(inputfilename.c_str());
         random_true = -1;
      }
      else
      {
         Coincidences = new TChain("Coincidences");
         Coincidences->Add(inputfilename.c_str());
         random_true = 1;
      }
     
      Coincidences->SetBranchStatus("*",0);
      Coincidences->SetBranchAddress("axialPos",&axialPos);
      Coincidences->SetBranchAddress("comptVolName1",&comptVolName1);
      Coincidences->SetBranchAddress("comptVolName2",&comptVolName2);
      Coincidences->SetBranchAddress("comptonCrystal1",&compton1);
      Coincidences->SetBranchAddress("comptonCrystal2",&compton2);
      Coincidences->SetBranchAddress("crystalID1",&crystalID1);
      Coincidences->SetBranchAddress("crystalID2",&crystalID2);
      Coincidences->SetBranchAddress("comptonPhantom1",&comptonPhantom1);
      Coincidences->SetBranchAddress("comptonPhantom2",&comptonPhantom2);
      Coincidences->SetBranchAddress("energy1",&energy1);
      Coincidences->SetBranchAddress("energy2",&energy2);    
      Coincidences->SetBranchAddress("eventID1",&eventID1);
      Coincidences->SetBranchAddress("eventID2",&eventID2);
      Coincidences->SetBranchAddress("globalPosX1",&globalPosX1);
      Coincidences->SetBranchAddress("globalPosX2",&globalPosX2);
      Coincidences->SetBranchAddress("globalPosY1",&globalPosY1);
      Coincidences->SetBranchAddress("globalPosY2",&globalPosY2);
      Coincidences->SetBranchAddress("globalPosZ1",&globalPosZ1);
      Coincidences->SetBranchAddress("globalPosZ2",&globalPosZ2);
      Coincidences->SetBranchAddress("layerID1",&layerID1);
      Coincidences->SetBranchAddress("layerID2",&layerID2);
      Coincidences->SetBranchAddress("moduleID1",&moduleID1);
      Coincidences->SetBranchAddress("moduleID2",&moduleID2);
      Coincidences->SetBranchAddress("rotationAngle",&rotationAngle);
      Coincidences->SetBranchAddress("rsectorID1",&rsectorID1);
      Coincidences->SetBranchAddress("rsectorID2",&rsectorID2);
      Coincidences->SetBranchAddress("runID",&runID);
      Coincidences->SetBranchAddress("sinogramS",&sinogramS);
      Coincidences->SetBranchAddress("sinogramTheta",&sinogramTheta);
      Coincidences->SetBranchAddress("sourceID1",&sourceID1);
      Coincidences->SetBranchAddress("sourceID2",&sourceID2);
      Coincidences->SetBranchAddress("sourcePosX1",&sourcePosX1);
      Coincidences->SetBranchAddress("sourcePosX2",&sourcePosX2);
      Coincidences->SetBranchAddress("sourcePosY1",&sourcePosY1);
      Coincidences->SetBranchAddress("sourcePosY2",&sourcePosY2);
      Coincidences->SetBranchAddress("sourcePosZ1",&sourcePosZ1);
      Coincidences->SetBranchAddress("sourcePosZ2",&sourcePosZ2);
      Coincidences->SetBranchAddress("submoduleID1",&submoduleID1);
      Coincidences->SetBranchAddress("submoduleID2",&submoduleID2);
      Coincidences->SetBranchAddress("time1",&time1);
      Coincidences->SetBranchAddress("time2",&time2);
         
      //#####################################################################
      //#   SINOGRAMS AND PROJECTION PLANES BINNING                         #
      //#####################################################################
      nentries = (Int_t)(Coincidences->GetEntries());

      if (t_cd == 0) printf("Total Number of Unfiltered Coincidence Events in File:= %d \n",nentries ); 
      if (t_cd == 1) printf("Total Number of Random Coincidence Events in File:= %d \n",nentries ); 
      
      for (Int_t i = 0 ; i < nentries ; i++)
      {
         if ((i%250000) == 0 && i!=0)   printf("... %d ",i);           
         if ((i%1000000) == 0 && i!=0)  printf("... %d\n",i);           
         
         nbytes += Coincidences->GetEntry(i);
         if (scan_id > -1 && runID != scan_id) continue;

         //-----------------------------------
         //   Identify the ring#...
         //-----------------------
         ring1 = (Int_t)(crystalID1/N_CRY_xy) 
         + (Int_t)(submoduleID1/N_SMOD_xy)*N_CRY_z
         + (Int_t)(moduleID1/N_MOD_xy)*N_SMOD_z*N_CRY_z;
         ring2 = (Int_t)(crystalID2/N_CRY_xy) 
         + (Int_t)(submoduleID2/N_SMOD_xy)*N_CRY_z
         + (Int_t)(moduleID2/N_MOD_xy)*N_SMOD_z*N_CRY_z;

         if ( abs(ring1 - ring2) > n_ring_dif ) { continue; }

         //-----------------------
         //   Identify the crystal#...
         //-----------------------------------
         crystal1 = rsectorID1 * N_MOD_xy * N_SMOD_xy * N_CRY_xy
            + (moduleID1%N_MOD_xy) * N_SMOD_xy * N_CRY_xy
            + (submoduleID1%N_SMOD_xy) * N_CRY_xy
            + (crystalID1%N_CRY_xy);
         crystal2 = rsectorID2 * N_MOD_xy * N_SMOD_xy * N_CRY_xy
            + (moduleID2%N_MOD_xy) * N_SMOD_xy * N_CRY_xy
            + (submoduleID2%N_SMOD_xy) * N_CRY_xy
            + (crystalID2%N_CRY_xy);

         //-----------------------------------------------------
         //   Rotate the image correctly#...
         //--------------------------------
         if (USE_OFFSET == 1)
         {
            crystal1 = crystal1 + OFFSET;
            crystal2 = crystal2 + OFFSET;      
            if (crystal1 >= N_DET) crystal1 = crystal1 - N_DET;
            if (crystal2 >= N_DET) crystal2 = crystal2 - N_DET;
         }
         //--------------------------------
         //   Bin the crystal ring pairs into Michelograms
         //   u - radial sinogram component
         //   phi - azimuthal sinogram component
         //   ring pairs are sorted according to c1 < c2 else flip
         //   where c1 and c2 are crystals at phi(u = S_WIDTH/2)
         //--------------------------------
         phi = ((crystal1 + crystal2 + N_DET/2)%N_DET)/2;
      
         if (((crystal1 + crystal2) < (3*N_DET/2)) && ((crystal1 + crystal2) >= (N_DET/2)))
            u = abs(crystal1 - crystal2) - N_DET/2 + S_WIDTH/2;
         else u = -abs(crystal1 - crystal2) + N_DET/2 + S_WIDTH/2;              

         if ( u >= S_WIDTH || u < 0 ) { continue; };

         if (u%2 == 0) 
         {
            zi = (N_DET/2 - (crystal1 - crystal2) - 1)/2;
            if (zi >= N_DET/4) zi = zi - N_DET/2 + 1;
            if (zi <= -N_DET/4) zi = zi + N_DET/2 - 1;
         }
         else                
         {
            zi = (N_DET/2 - (crystal1 - crystal2))/2;
            if (zi >= N_DET/4) zi = zi - N_DET/2;
            if (zi <= -N_DET/4) zi = zi + N_DET/2;
         }

         c1 = crystal1 + zi;
         c2 = crystal2 - zi;
         if (c1 >= N_DET) c1 = c1 - N_DET;
         if (c1 < 0) c1 = c1 + N_DET;
         if (c2 >= N_DET) c2 = c2 - N_DET;
         if (c2 < 0) c2 = c2 + N_DET;
         
         // Flip
         if (c1 < c2) ring1 = (ring1+ring2) - (ring2=ring1);

         // Update the different arrays...
         //-------------------------------
         if (eventID1 == eventID2)
         {
            if (comptonPhantom1 == 0 && comptonPhantom1 == 0)
            {
               if (b_trues == 1) 
               {
                  Mich_r1r2fu[ring2][ring1][phi][u] += random_true;
                  Counts += random_true;
               }
            }
            else if (b_scatters == 1)
            {
               Mich_r1r2fu[ring2][ring1][phi][u] += random_true;
               Counts += random_true;
            }
         }
         else if (b_randoms == 1)
         {
            Mich_r1r2fu[ring2][ring1][phi][u] += random_true;
            Counts += random_true;
         }
      }
   }
   printf("Total Number of Filtered Coincidence Events in Scan:= %d \n",Counts ); 
   printf("\n");

   // Write the data to disk, and then close Michelogram file...
   //----------------------------------------------------------------
   fwrite(Mich_r1r2fu,4,(N_RINGS*N_RINGS*N_DET/2*S_WIDTH),Mich_r1r2fuFile);
   fclose(Mich_r1r2fuFile);
   
   // Generate projection files...
   //-----------------------------
   // From Segment number -n_ring_dif to +n_ring_dif
   // After phi
   // After z
   // After r
   Int_t S_NUM;
   for (int i = 0 ; i < 2*n_ring_dif + 1 ; i++)
   {
      if (i <= n_ring_dif) S_NUM = N_RINGS - n_ring_dif + i;
      else S_NUM = N_RINGS + n_ring_dif - i;
      for (int j = 0 ; j < N_DET/2 ; j++)
      {
         float Proj[S_NUM][S_WIDTH];
         for (int k = 0 ; k < S_NUM ; k++)
         {
            if (i <= n_ring_dif) ring1 = k;
            else ring2 = k;
            if (i <= n_ring_dif) ring2 = ring1 + n_ring_dif - i;
            else ring1 = ring2 - n_ring_dif + i;
               for (int l = 0 ; l < S_WIDTH ; l++) Proj[k][l] = Mich_r1r2fu[ring2][ring1][j][l];
         }
         fwrite(Proj,4,(S_NUM*S_WIDTH),Proj_File); 
      }
   }
      
   fclose(Proj_File);
   //-----------------------------
   
   return(0);
}
